/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.build.service;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.Getter;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.build.BuildProps;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.build.GeneralBuild;
import org.artifactory.api.build.ModuleArtifact;
import org.artifactory.api.build.ModuleDependency;
import org.artifactory.api.build.PublishedModule;
import org.artifactory.api.build.model.diff.BuildParams;
import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiBuild;
import org.artifactory.aql.api.internal.AqlApiDynamicFieldsDomains;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.result.rows.AqlBaseFullRowImpl;
import org.artifactory.build.BuildId;
import org.artifactory.build.BuildInfoUtils;
import org.artifactory.build.BuildNameAndRepo;
import org.artifactory.build.BuildRun;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.common.ConstantValues;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.binstore.service.BinaryService;
import org.artifactory.storage.build.service.BuildSearchCriteria;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.build.dao.BuildArtifactsDao;
import org.artifactory.storage.db.build.dao.BuildDependenciesDao;
import org.artifactory.storage.db.build.dao.BuildModulesDao;
import org.artifactory.storage.db.build.dao.BuildsDao;
import org.artifactory.storage.db.build.entity.BuildArtifact;
import org.artifactory.storage.db.build.entity.BuildDependency;
import org.artifactory.storage.db.build.entity.BuildEntity;
import org.artifactory.storage.db.build.entity.BuildIdEntity;
import org.artifactory.storage.db.build.entity.BuildModule;
import org.artifactory.storage.db.build.entity.BuildPromotionStatus;
import org.artifactory.storage.db.build.entity.BuildProperty;
import org.artifactory.storage.db.build.entity.ModuleProperty;
import org.artifactory.storage.db.util.blob.BlobWrapperFactory;
import org.artifactory.storage.jobs.migration.buildinfo.BuildMigrationUtils;
import org.artifactory.ui.rest.service.builds.search.BuildsSearchFilter;
import org.artifactory.util.CollectionUtils;
import org.jfrog.build.api.Artifact;
import org.jfrog.build.api.Build;
import org.jfrog.build.api.BuildFileBean;
import org.jfrog.build.api.Dependency;
import org.jfrog.build.api.release.PromotionStatus;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.StreamSupportUtils;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.wrapper.BlobWrapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nullable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Properties;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.lang.String.format;
import static org.artifactory.aql.api.domain.sensitive.AqlApiBuild.buildRepo;
import static org.artifactory.aql.api.domain.sensitive.AqlApiBuild.name;
import static org.artifactory.aql.api.domain.sensitive.AqlApiBuild.number;
import static org.artifactory.aql.api.domain.sensitive.AqlApiBuild.started;
import static org.artifactory.build.BuildInfoUtils.parseBuildTime;

/**
 * Date: 11/14/12
 * Time: 12:42 PM
 *
 * @author freds
 */
@Service
@Slf4j
public class BuildStoreServiceImpl implements InternalBuildStoreService {

    private InternalDbService dbService;
    private BinaryService binaryService;
    private BuildsDao buildsDao;
    private BuildModulesDao buildModulesDao;
    private BuildArtifactsDao buildArtifactsDao;
    private BuildDependenciesDao buildDependenciesDao;
    private BlobWrapperFactory blobsFactory;
    private AqlService aqlService;

    public static Date parseStringToDate(String dateString) {
        return new Date(parseBuildTime(dateString));
    }

    public static String formatDateToString(long buildStarted) {
        return BuildInfoUtils.formatBuildTime(buildStarted);
    }

    @Override
    public void addBuild(@NonNull BuildWrapper build, boolean stateReady) {
        try {
            String buildStarted = build.getStarted();
            Date parsedDate = parseStringToDate(buildStarted);

            // TODO: [by fsi] we are loosing the timezone information written in the JSON
            // Generates a big inconsistency between DB entry and JSON data
            BuildEntity dbBuild = new BuildEntity(dbService.nextId(), build.getName(), build.getNumber(),
                    parsedDate.getTime(),
                    build.getBuild().getUrl(), System.currentTimeMillis(), build.getBuild().getArtifactoryPrincipal(),
                    0L, null, build.getBuildRepo());
            long buildId = dbBuild.getBuildId();
            dbBuild.setProperties(createProperties(buildId, build.getBuild()));
            dbBuild.setPromotions(createPromotions(buildId, build.getBuild()));
            BlobWrapper buildJson = getBuildJsonWrapper(build.getBuild(), stateReady);
            buildsDao.createBuild(dbBuild, buildJson);
            insertModules(buildId, build.getBuild());
        } catch (SQLException e) {
            throw new StorageException("Could not insert build " + build, e);
        }
    }

    private ArrayList<BuildPromotionStatus> createPromotions(long buildId, Build build) {
        List<PromotionStatus> statuses = build.getStatuses();
        ArrayList<BuildPromotionStatus> buildPromotions;
        if (statuses != null && !statuses.isEmpty()) {
            buildPromotions = new ArrayList<>(statuses.size());
            for (PromotionStatus status : statuses) {
                buildPromotions.add(convertPromotionStatus(buildId, status));
            }
        } else {
            buildPromotions = new ArrayList<>(1);
        }
        return buildPromotions;
    }

    private BuildPromotionStatus convertPromotionStatus(long buildId, PromotionStatus status) {
        return new BuildPromotionStatus(buildId,
                status.getTimestampDate().getTime(),
                status.getUser(),
                status.getStatus(),
                status.getRepository(),
                status.getComment(),
                status.getCiUser());
    }

    private Set<BuildProperty> createProperties(long buildId, Build build) {
        Properties properties = build.getProperties();
        Set<BuildProperty> buildProperties;
        if (properties != null && !properties.isEmpty()) {
            buildProperties = new HashSet<>(properties.size());
            for (Map.Entry<Object, Object> entry : properties.entrySet()) {
                buildProperties.add(
                        new BuildProperty(dbService.nextId(), buildId, entry.getKey().toString(),
                                entry.getValue().toString())
                );
            }
        } else {
            buildProperties = new HashSet<>(1);
        }
        return buildProperties;
    }

    private void insertModules(long buildId, Build build) throws SQLException {
        List<org.jfrog.build.api.Module> modules = build.getModules();
        if (modules == null || modules.isEmpty()) {
            // Nothing to do here
            return;
        }
        for (org.jfrog.build.api.Module module : modules) {
            BuildModule dbModule = new BuildModule(dbService.nextId(), buildId, module.getId());
            Properties properties = module.getProperties();
            dbModule.setProperties(calculateModuleProperties(dbModule, properties));
            buildModulesDao.createBuildModule(dbModule);

            List<Artifact> artifacts = module.getArtifacts();
            buildArtifactsDao.createBuildArtifacts(calculateBuildArtifacts(dbModule, artifacts));

            List<Dependency> dependencies = module.getDependencies();
            List<BuildDependency> dbDependencies;
            if (dependencies != null && !dependencies.isEmpty()) {
                dbDependencies = Lists.newArrayListWithExpectedSize(dependencies.size());
                for (Dependency dependency : dependencies) {
                    // Dependency properties are not inserted in DB
                    dbDependencies.add(new BuildDependency(dbService.nextId(), dbModule.getModuleId(),
                            dependency.getId(), dependency.getScopes(), dependency.getType(),
                            dependency.getSha1(), dependency.getMd5(), dependency.getSha256()));
                }
            } else {
                dbDependencies = Lists.newArrayListWithExpectedSize(1);
            }
            buildDependenciesDao.createBuildDependencies(dbDependencies);
        }
    }

    @NonNull
    private List<BuildArtifact> calculateBuildArtifacts(@NonNull BuildModule dbModule, List<Artifact> artifacts) {
        List<BuildArtifact> dbArtifacts;
        if (artifacts != null && !artifacts.isEmpty()) {
            dbArtifacts = Lists.newArrayListWithExpectedSize(artifacts.size());
            for (Artifact artifact : artifacts) {
                // Artifact properties are not inserted in DB
                dbArtifacts.add(new BuildArtifact(dbService.nextId(), dbModule.getModuleId(),
                        artifact.getName(), artifact.getType(), artifact.getSha1(), artifact.getMd5(),
                        artifact.getSha256()));
            }
        } else {
            dbArtifacts = Lists.newArrayListWithExpectedSize(1);
        }
        return dbArtifacts;
    }

    @NonNull
    private Set<ModuleProperty> calculateModuleProperties(@NonNull BuildModule dbModule, Properties properties) {
        Set<ModuleProperty> moduleProperties;
        if (properties != null && !properties.isEmpty()) {
            moduleProperties = Sets.newHashSetWithExpectedSize(properties.size());
            for (Map.Entry<Object, Object> entry : properties.entrySet()) {
                moduleProperties.add(
                        new ModuleProperty(dbService.nextId(), dbModule.getModuleId(),
                                entry.getKey().toString(),
                                entry.getValue().toString())
                );
            }
        } else {
            moduleProperties = Sets.newHashSetWithExpectedSize(1);
        }
        return moduleProperties;
    }

    @Override
    public void populateMissingChecksums(Build build) {
        List<org.jfrog.build.api.Module> modules = build.getModules();
        if (modules != null && !modules.isEmpty()) {
            for (org.jfrog.build.api.Module module : modules) {
                handleBeanPopulation(module.getArtifacts());
                handleBeanPopulation(module.getDependencies());
            }
        }
    }

    @Override
    public BuildRun getBuildRun(String buildName, String buildNumber, String buildStarted) {
        try {
            Date parsedDate = parseStringToDate(buildStarted);
            BuildEntity entity = buildsDao.findBuild(buildName, buildNumber, parsedDate.getTime());
            if (entity != null) {
                return getBuildRun(entity);
            }
        } catch (SQLException e) {
            throw new StorageException("Could not execute find build for build" +
                    " name='" + buildName + "' number='" + buildNumber + "' start='" + buildStarted + "'", e);
        }
        return null;
    }

    @Override
    public void addPromotionStatus(BuildWrapper build, PromotionStatus promotion, String currentUser,
            boolean stateReady) {
        if (log.isDebugEnabled()) {
            log.debug("Adding build promotion status for build '{}'", build);
        }
        BuildRun buildRun = getBuildRun(build.getName(), build.getNumber(), build.getStarted());
        if (buildRun == null) {
            throw new StorageException("Could not add promotion " + promotion + " to non existent build " + build);
        }
        try {
            long buildId = findIdFromBuildRun(buildRun);
            BlobWrapper buildJson = getBuildJsonWrapper(build.getBuild(), stateReady);
            buildsDao.addPromotionStatus(buildId, convertPromotionStatus(buildId, promotion),
                    buildJson, currentUser, System.currentTimeMillis());
        } catch (SQLException e) {
            throw new StorageException("Could not add promotion " + promotion + " for build " + buildRun, e);
        }
    }

    @Override
    public boolean exists(BuildWrapper build) {
        String name = build.getName();
        String number = build.getNumber();
        String started = build.getStarted();
        try {
            return buildsDao.findBuildId(name, number, parseBuildTime(started)) > 0;
        } catch (SQLException e) {
            throw new StorageException(
                    format("Error encountered while verifying build existence for build %s:%s-%s", name, number,
                            started), e);
        }
    }

    @Override
    public Build getBuild(BuildRun buildRun) {
        try {
            boolean shouldFixProperties = ConstantValues.buildInfoMigrationFixProperties.getBoolean();
            long buildId = findIdFromBuildRun(buildRun);
            if (buildId > 0L) {
                return shouldFixProperties ? JsonUtils.getInstance()
                        .readValue(
                                BuildMigrationUtils.fixBuildProperties(buildsDao.getJsonBuild(buildId, String.class)),
                                Build.class) :
                        buildsDao.getJsonBuild(buildId, Build.class);
            }
        } catch (SQLException e) {
            throw new StorageException("Could not execute get build JSON for build " + buildRun, e);
        }
        return null;
    }

    @Override
    public BuildRun getLatestBuildRun(String buildName, String buildNumber, String repoKey) {
        String fullBuildDate = "not found";
        try {
            return buildsDao.findLatestBuild(buildName, buildNumber, repoKey);
        } catch (SQLException e) {
            throw new StorageException("Could not find build JSON for latest build" +
                    " name='" + buildName + "' number='" + buildNumber + "' latest date found='" + fullBuildDate + "'",
                    e
            );
        }
    }

    @Override
    public BuildEntity getBuildEntity(BuildRun buildRun) {
        try {
            BuildEntity buildEntity =
                    buildsDao.findBuild(buildRun.getName(), buildRun.getNumber(), buildRun.getStartedDate().getTime());
            if (buildEntity == null) {
                throw new StorageException("Cannot create exportable build of non existent build " + buildRun);
            }
            return buildEntity;
        } catch (SQLException e) {
            throw new StorageException("Could not create exportable build object for " + buildRun, e);
        }
    }

    private void deleteBuilds(Collection<Long> buildIds, boolean stateReady) throws SQLException {
        for (Long buildId : buildIds) {
            List<Long> moduleIds = buildModulesDao.findModuleIdsForBuild(buildId);
            if (!moduleIds.isEmpty()) {
                buildArtifactsDao.deleteBuildArtifacts(moduleIds);
                buildDependenciesDao.deleteBuildDependencies(moduleIds);
            }
            buildModulesDao.deleteBuildModules(moduleIds, buildId);
            buildsDao.deleteBuild(buildId, !stateReady);
        }
    }

    @Override
    public void deleteBuild(String buildName, String buildNumber, String buildStarted, String repo,
            boolean stateReady) {
        BuildRunImpl buildRun = new BuildRunImpl(buildName, buildNumber, buildStarted, repo);
        log.debug("Deleting Build " + buildRun);
        try {
            long buildId = findIdFromBuildRun(buildRun);
            if (buildId > 0L) {
                deleteBuilds(ImmutableList.of(buildId), stateReady);
            } else {
                log.info("Build " + buildRun + " already deleted!");
            }
        } catch (SQLException e) {
            throw new StorageException("Could not delete build " + buildRun, e);
        }
    }

    @Override
    public void deleteAllBuilds(boolean stateReady) {
        try {
            buildArtifactsDao.deleteAllBuildArtifacts();
            buildDependenciesDao.deleteAllBuildDependencies();
            buildModulesDao.deleteAllBuildModules();
            buildsDao.deleteAllBuilds(!stateReady);
        } catch (SQLException e) {
            throw new StorageException("Could not delete all builds", e);
        }
    }

    @Override
    public Set<BuildRun> findBuildsForChecksum(BuildSearchCriteria criteria, ChecksumType type, String checksum,
            @NonNull String buildRepo) {
        if (!type.isValid(checksum)) {
            log.info("Looking for invalid checksum {} '{}'", type.name(), checksum);
        }
        try {
            Set<BuildRun> results = Sets.newHashSet();
            if (criteria.searchInDependencies()) {
                Collection<BuildEntity> buildEntities =
                        buildsDao.findBuildsForDependencyChecksum(type, checksum, buildRepo);
                for (BuildEntity buildEntity : buildEntities) {
                    results.add(getBuildRun(buildEntity));
                }
            }

            if (criteria.searchInArtifacts()) {
                Collection<BuildEntity> buildEntities =
                        buildsDao.findBuildsForArtifactChecksum(type, checksum, buildRepo);
                for (BuildEntity buildEntity : buildEntities) {
                    results.add(getBuildRun(buildEntity));
                }
            }

            return results;
        } catch (SQLException e) {
            throw new StorageException("Could not find builds for " + type.name() + " '" + checksum + "'", e);
        }
    }

    @Override
    public List<PublishedModule> getPublishedModules(String buildNumber, String date, String buildRepo) {
        List<PublishedModule> buildModule = null;
        try {
            buildModule = buildsDao.getBuildModule(buildNumber, date, buildRepo);
        } catch (Exception e) {
            log.error(e.toString());
        }
        return buildModule;
    }

    @Override
    public List<ModuleArtifact> getModuleArtifact(String buildName, String buildNumber, String moduleId, String date,
            String buildRepo, boolean onlyDefaultBuildRepoExists) {
        List<ModuleArtifact> moduleArtifactList = null;
        try {
            moduleArtifactList = buildsDao
                    .getModuleArtifact(buildName, buildNumber, date, moduleId, buildRepo, onlyDefaultBuildRepoExists);
        } catch (Exception e) {
            log.error(e.toString());
        }
        return moduleArtifactList;
    }

    @Override
    public List<ModuleDependency> getModuleDependency(String buildNumber, String moduleId, String date,
            String buildRepo) {
        List<ModuleDependency> moduleArtifactList = null;
        try {
            moduleArtifactList = buildsDao.getModuleDependency(buildNumber, date, buildRepo, moduleId);
        } catch (Exception e) {
            log.error(e.toString());
        }
        return moduleArtifactList;
    }

    @Override
    public List<ModuleArtifact> getModuleArtifactsForDiffWithPaging(BuildParams buildParams,
            boolean onlyDefaultBuildRepoExists) {
        return buildsDao.getModuleArtifactsForDiffWithPaging(buildParams, onlyDefaultBuildRepoExists);
    }

    public List<GeneralBuild> getPrevBuildsList(String buildName, String buildDate, String buildRepo) {
        return buildsDao.getPrevBuildsList(buildName, buildDate, buildRepo);
    }

    @Override
    public List<BuildProps> getBuildPropsData(BuildParams buildParams) {
        List<BuildProps> buildPropsList = buildsDao.getBuildPropsList(buildParams);
        return distinctBuildProps(buildPropsList);
    }

    public static List<BuildProps> distinctBuildProps(List<BuildProps> buildPropertyList) {
        List<BuildPropertyAsKeyValue> distinct = buildPropertyList.stream()
                .map(BuildPropertyAsKeyValue::new).distinct().collect(Collectors.toList());

        return distinct.stream().map(it -> new BuildProps(it.propKey, it.propValue, null)).collect(Collectors.toList());
    }

    @Override
    public List<ModuleDependency> getModuleDependencyForDiffWithPaging(BuildParams buildParams) {
        return buildsDao.getModuleDependencyForDiffWithPaging(buildParams);
    }

    @Override
    public Set<BuildRun> getLatestBuildsByName(String buildRepo) {
        try {
            List<BuildNameAndRepo> allBuildsInfo = buildsDao.getAllBuildNames(buildRepo);
            LinkedHashSet<BuildRun> results = new LinkedHashSet<>(allBuildsInfo.size());
            for (BuildNameAndRepo buildInfo : allBuildsInfo) {
                BuildEntity buildEntity = buildsDao.getLatestBuild(buildInfo.getName(), buildInfo.getRepo());
                if (buildEntity != null) {
                    results.add(getBuildRun(buildEntity));
                }
            }
            return results;
        } catch (SQLException e) {
            throw new StorageException("Could not list all builds by name and latest build date", e);
        }
    }

    @Override
    public List<BuildId> getLatestBuildIDsPaging(ContinueBuildFilter continueBuildFilter) {
        List<BuildIdEntity> allBuildNames = buildsDao.getLatestBuildIds(continueBuildFilter);
        return StreamSupportUtils.stream(allBuildNames)
                .map(this::buildIDFromBuildIDEntity)
                .collect(Collectors.toList());
    }

    @Override
    public List<BuildId> findLatestBuildIDsByNameCaseInsensitive(String buildRepo, String buildName,
            String fromDate, String toDate,
            String orderBy, String direction) {
        try {
            List<BuildIdEntity> builds = buildsDao
                    .findLatestBuildIdsByNameCaseInsensitive(buildRepo, buildName, fromDate, toDate, orderBy,
                            direction);
            return builds.stream()
                    .map(this::buildIDFromBuildIDEntity)
                    .collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Could not list all builds by name and latest build date", e);
        }
    }

    /**
     * The AQL query is sorting according to user's filter + Promotion date.
     * The AQL results are sorted and then the list is being streamed with {@link Stream#distinct()} for getting
     * only the first of each build, which is actually the latest promoted status.
     */
    @Override
    public List<GeneralBuild> getBuildNumbersByBuildName(BuildsSearchFilter filter) {
        List<AqlBaseFullRowImpl> allBuildNames = getAqlBuildNumbersByBuildName(filter);

        return allBuildNames.stream()
                .map(this::createGeneralInfo)
                .distinct() // As the results contain multiple records per Build (due to db record per promotion status)
                // the first build will be the one with latest promotion status
                .collect(Collectors.toList());
    }

    private GeneralBuild createGeneralInfo(AqlBaseFullRowImpl build) {
        return new GeneralBuild(build.getBuildName(), build.getBuildNumber(), build.getBuildStarted().getTime(),
                build.getBuildUrl(), build.getBuildPromotionStatus(), build.getBuildRepo());
    }

    /**
     * @return A list of sorted results according to user's filter & promotion date
     */
    private List<AqlBaseFullRowImpl> getAqlBuildNumbersByBuildName(BuildsSearchFilter filter) {
        AqlBase.CriteriaClause<AqlApiBuild> buildRepoEquals = null;
        AqlBase.CriteriaClause<AqlApiBuild> nameEquals = null;
        AqlBase.CriteriaClause<AqlApiBuild> before = null;
        AqlBase.CriteriaClause<AqlApiBuild> after = null;

        AqlBase.OrClause<AqlApiBuild> numbersOrClause = null;

        if (StringUtils.isNotEmpty(filter.getBuildRepo())) {
            buildRepoEquals = buildRepo().equal(filter.getBuildRepo());
        }
        if (StringUtils.isNotBlank(filter.getName())) {
            nameEquals = name().equal(filter.getName());
        }
        if (filter.getAfter() != 0) {
            after = started().greaterEquals(filter.getAfter());
        }
        if (filter.getBefore() != 0) {
            before = started().lessEquals(filter.getBefore());
        }
        if (filter.getNumbers() != null) {
            numbersOrClause = AqlBase.or();
            for (String number : filter.getNumbers()) {
                numbersOrClause.append(number().equal(number));
            }
        }

        AqlBase aql = AqlApiBuild.create().include(AqlApiBuild.number(), AqlApiBuild.url(), AqlApiBuild.started(),
                AqlApiBuild.buildPromotions().status(), AqlApiBuild.buildPromotions().created(),
                AqlApiBuild.buildRepo())
                .filter(AqlBase.and(
                        buildRepoEquals,
                        nameEquals,
                        after,
                        before,
                        numbersOrClause
                ))
                .addSortElement(sortResultsBy(filter))
                // Extra sort according to promotion date - as we eventually want only the latest
                .addSortElement(AqlApiBuild.buildPromotions().created()).desc()
                .desc();

        if (filter.getDirection().equals("asc")) {
            aql.asc();
        }

        return aqlService.executeQueryEager(aql).getResults();
    }

    private static AqlApiDynamicFieldsDomains.AqlApiComparator<AqlApiBuild> sortResultsBy(BuildsSearchFilter filter) {
        if ("number".equals(filter.getOrderBy())) {
            return number();
        }
        return AqlApiBuild.started(); // "build_date"
    }

    private BuildIdImpl buildIDFromBuildIDEntity(BuildIdEntity buildIDEntity) {
        return new BuildIdImpl(buildIDEntity.getBuildId(), buildIDEntity.getBuildName(),
                buildIDEntity.getBuildNumber(), buildIDEntity.getBuildDate(), buildIDEntity.getBuildRepo());
    }

    @Override
    public Set<BuildRun> findBuildsByName(String buildName, @NonNull String buildRepo) {
        Set<BuildRun> results = Sets.newHashSet();
        try {
            List<Long> buildIds = buildsDao.findBuildIds(buildName, buildRepo);
            for (Long buildId : buildIds) {
                getBuildRun(buildId).ifPresent(results::add);
            }
        } catch (SQLException e) {
            throw new StorageException("Could not search for builds with name='" + buildName + "'", e);
        }
        return results;
    }

    @Override
    public Set<BuildRun> findBuildsByNameAndNumber(String buildName, String buildNumber, @NonNull String buildRepo) {
        Set<BuildRun> results = Sets.newHashSet();
        try {
            List<Long> buildIds = buildsDao.findBuildIds(buildName, buildNumber, buildRepo);
            for (Long buildId : buildIds) {
                getBuildRun(buildId).ifPresent(results::add);
            }
        } catch (SQLException e) {
            throw new StorageException(
                    "Could not search for builds with name='" + buildName + "' and number='" + buildNumber + "'", e);
        }
        return results;
    }

    @Override
    public List<BuildNameAndRepo> getAllBuildsInfo(@Nullable String buildRepo) {
        try {
            return buildsDao.getAllBuildNames(buildRepo);
        } catch (SQLException e) {
            throw new StorageException("Could not retrieve the list of build names", e);
        }
    }

    private long findIdFromBuildRun(BuildRun buildRun) throws SQLException {
        long buildId = 0L;
        if (buildRun instanceof BuildRunImpl) {
            buildId = ((BuildRunImpl) buildRun).getBuildId();
        }
        if (buildId <= 0L) {
            buildId = buildsDao
                    .findBuildId(buildRun.getName(), buildRun.getNumber(), buildRun.getStartedDate().getTime());
        }
        return buildId;
    }

    @Override
    public long findIdFromBuild(BuildRun build) {
        try {
            return buildsDao.findBuildId(build.getName(), build.getNumber(), parseBuildTime(build.getStarted()));
        } catch (SQLException e) {
            throw new StorageException("Could not get build id by name, number and started date", e);
        }
    }

    @NonNull
    private Optional<BuildRun> getBuildRun(Long buildId) throws SQLException {
        BuildEntity buildEntity = buildsDao.getBuild(buildId);
        if (buildEntity == null) {
            return Optional.empty();
        }
        return Optional.of(getBuildRun(buildEntity));
    }

    @Override
    public List<GeneralBuild> getBuildForName(String buildName, String buildRepo,
            ContinueBuildFilter continueBuildFilter) {
        try {
            return buildsDao.getBuildForName(buildName, buildRepo, continueBuildFilter);
        } catch (SQLException e) {
            throw new StorageException(String.format("Could not get builds for name '%s'.", buildName), e);
        }
    }

    private BuildRun getBuildRun(BuildEntity buildEntity) {
        String releaseStatus = null;
        if (!buildEntity.getPromotions().isEmpty()) {
            releaseStatus = buildEntity.getPromotions().last().getStatus();
        }
        return new BuildRunImpl(buildEntity.getBuildId(), buildEntity.getBuildName(), buildEntity.getBuildNumber(),
                formatDateToString(buildEntity.getBuildDate()), buildEntity.getCiUrl(), releaseStatus,
                buildEntity.getRepo());
    }

    @Nullable
    private BlobWrapper getBuildJsonWrapper(Build build, boolean stateReady) {
        // Fully migrated to build-info repo, no need to save the build json in db
        BlobWrapper jsonObject = null;
        if (!stateReady) {
            // Legacy code for an instance with upgrade migration not yet complete
            jsonObject = blobsFactory.createJsonObjectWrapper(build);
        }
        return jsonObject;
    }

    /**
     * Locates and fills in missing checksums of a build file bean
     *
     * @param buildFiles List of build files to populate
     */
    private void handleBeanPopulation(List<? extends BuildFileBean> buildFiles) {
        if (CollectionUtils.isNullOrEmpty(buildFiles)) {
            return;
        }
        Set<String> checksums = getMissingChecksums(buildFiles);

        Set<BinaryElement> binaryInfos = binaryService.findBinaries(checksums);
        Map<String, BinaryElement> sha1ToInfo = Maps.newHashMap();
        Map<String, BinaryElement> sha2ToInfo = Maps.newHashMap();
        Map<String, BinaryElement> md5ToInfo = Maps.newHashMap();

        binaryInfos.forEach(binaryInfo -> {
            sha1ToInfo.put(binaryInfo.getSha1(), binaryInfo);
            sha2ToInfo.put(binaryInfo.getSha2(), binaryInfo);
            md5ToInfo.put(binaryInfo.getMd5(), binaryInfo);
        });
        populateMissingChecksums(buildFiles, sha1ToInfo, sha2ToInfo, md5ToInfo);
    }

    /**
     * Collects one available checksum for each {@link BuildFileBean} in {@param buildFiles} that has one or more
     * missing checksums from its build info.
     * If all checksums are missing from the build info or if the bean has all, it is skipped.
     */
    private Set<String> getMissingChecksums(List<? extends BuildFileBean> buildFiles) {
        Set<String> checksums = Sets.newHashSet();
        for (BuildFileBean buildFile : buildFiles) {
            ShaExtractor shaExtractor = new ShaExtractor(buildFile);
            if (shaExtractor.skipProcessing()) {
                continue;
            }

            //This bean is missing one or more checksums - need to search for the others based on the first one it has.
            if (shaExtractor.isSha1Exists()) {
                checksums.add(buildFile.getSha1());
            } else if (shaExtractor.isSha2Exists()) {
                checksums.add(buildFile.getSha256());
            } else {
                checksums.add(buildFile.getMd5());
            }
        }
        return checksums;
    }

    private void populateMissingChecksums(List<? extends BuildFileBean> buildFiles,
            Map<String, BinaryElement> sha1ToInfo, Map<String, BinaryElement> sha2ToInfo,
            Map<String, BinaryElement> md5ToInfo) {

        for (BuildFileBean buildFile : buildFiles) {
            setMissingChecksums(sha1ToInfo, sha2ToInfo, md5ToInfo, buildFile);
        }
    }

    private void setMissingChecksums(Map<String, BinaryElement> sha1ToInfo, Map<String, BinaryElement> sha2ToInfo,
            Map<String, BinaryElement> md5ToInfo, BuildFileBean buildFile) {
        ShaExtractor shaExtractor = new ShaExtractor(buildFile);
        //If the bean has all or none of the checksums, skip it
        if (shaExtractor.skipProcessing()) {
            return;
        }
        if (shaExtractor.isSha1Exists()) {
            BinaryElement info = sha1ToInfo.get(buildFile.getSha1());
            if (info == null) {
                return;
            }
            if (!shaExtractor.isSha2Exists()) {
                setNewChecksum(ChecksumType.sha256, info.getSha2(), buildFile::setSha256);
            }
            if (!shaExtractor.isMd5Exists()) {
                setNewChecksum(ChecksumType.md5, info.getMd5(), buildFile::setMd5);
            }
        } else if (shaExtractor.isSha2Exists()) {
            BinaryElement info = sha2ToInfo.get(buildFile.getSha256());
            if (info == null) {
                return;
            }
            //there's no sha1 here for sure
            setNewChecksum(ChecksumType.sha1, info.getSha1(), buildFile::setSha1);
            if (!shaExtractor.isMd5Exists()) {
                setNewChecksum(ChecksumType.md5, info.getMd5(), buildFile::setMd5);
            }
        } else { //md5
            BinaryElement info = md5ToInfo.get(buildFile.getMd5());
            if (info == null) {
                return;
            }
            //there's no sha1 and no sha2 here for sure
            setNewChecksum(ChecksumType.sha256, info.getSha2(), buildFile::setSha256);
            setNewChecksum(ChecksumType.sha1, info.getSha1(), buildFile::setSha1);
        }
    }

    private void setNewChecksum(ChecksumType checksumType, String newChecksum, Consumer<String> checksumSetter) {
        if (checksumType.isValid(newChecksum)) {
            checksumSetter.accept(newChecksum);
        }
    }

    @Override
    public String getBuildStoringRepo(@NonNull String name, @NonNull String number, @NonNull String started) {
        try {
            Date parsedDate = parseStringToDate(started);
            return buildsDao.getBuildStoringRepo(name, number, parsedDate.getTime());
        } catch (SQLException e) {
            throw new StorageException("Could not find build repo by name and number", e);
        }
    }

    @Override
    public int getBuildsCountFromSpecificRepo(@Nullable String buildInfoRepoKey) {
        try {
            return buildsDao.getBuildsCountFromSpecificRepo(buildInfoRepoKey);
        } catch (SQLException e) {
            throw new StorageException("Could not count the total builds amount from repository " + buildInfoRepoKey,
                    e);
        }
    }

    @Override
    public int getUniqueBuildsCountFromSpecificRepo(String buildInfoRepoKey) {
        try {
            return buildsDao.getUniqueBuildsCountFromSpecificRepo(buildInfoRepoKey);
        } catch (SQLException e) {
            throw new StorageException("Could not count the total builds amount from repository " + buildInfoRepoKey,
                    e);
        }
    }

    @Override
    public int getTotalBuildsCount() {
        try {
            return buildsDao.getTotalBuildsCount();
        } catch (SQLException e) {
            throw new StorageException("Could not count the total builds amount", e);
        }
    }

    @Autowired
    public void setDbService(InternalDbService dbService) {
        this.dbService = dbService;
    }

    @Autowired
    public void setBinaryService(BinaryService binaryService) {
        this.binaryService = binaryService;
    }

    @Autowired
    public void setBuildsDao(BuildsDao buildsDao) {
        this.buildsDao = buildsDao;
    }

    @Autowired
    public void setBuildModulesDao(BuildModulesDao buildModulesDao) {
        this.buildModulesDao = buildModulesDao;
    }

    @Autowired
    public void setBuildArtifactsDao(BuildArtifactsDao buildArtifactsDao) {
        this.buildArtifactsDao = buildArtifactsDao;
    }

    @Autowired
    public void setBuildDependenciesDao(BuildDependenciesDao buildDependenciesDao) {
        this.buildDependenciesDao = buildDependenciesDao;
    }

    @Autowired
    public void setBlobsFactory(BlobWrapperFactory blobsFactory) {
        this.blobsFactory = blobsFactory;
    }

    @Autowired
    public void setAqlService(AqlService aqlService) {
        this.aqlService = aqlService;
    }

    @Getter
    private static class ShaExtractor {
        private final boolean sha1Exists;
        private final boolean sha2Exists;
        private final boolean md5Exists;

        ShaExtractor(BuildFileBean buildFile) {
            sha1Exists = StringUtils.isNotBlank(buildFile.getSha1());
            sha2Exists = StringUtils.isNotBlank(buildFile.getSha256());
            md5Exists = StringUtils.isNotBlank(buildFile.getMd5());
        }

        boolean skipProcessing() {
            return ((sha1Exists && sha2Exists && md5Exists) || (!sha1Exists && !sha2Exists && !md5Exists));
        }
    }
}
