package org.artifactory.storage.jobs.async.database;

import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.storage.db.locks.service.DbLocksServiceImpl;
import org.artifactory.storage.db.version.converter.DBConverter;
import org.artifactory.util.CollectionUtils;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JdbcHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * @author Alexei Vainshtein
 */
public class DatabaseConverterRunnable implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(DatabaseConverterRunnable.class);
    private final ArtifactoryHome artifactoryHome;
    private final List<DBConverter> converters;
    private final JdbcHelper jdbcHelper;
    public static final String ARTIFACTORY_ASYNC_CONVERTER_LOCK_KEY = "artifactory_async_converter";

    public DatabaseConverterRunnable(ArtifactoryHome artifactoryHome, List<DBConverter> converters,
            JdbcHelper jdbcHelper) {
        this.artifactoryHome = artifactoryHome;
        this.converters = converters;
        this.jdbcHelper = jdbcHelper;
    }

    @Override
    public void run() {
        log.info("Starting Async converter thread.");
        if (CollectionUtils.isNullOrEmpty(converters)) {
            log.debug("No async database converters were scheduled to run.");
            return;
        }
        String ownerId = artifactoryHome.getHaNodeProperties().getServerId();
        ArtifactoryDbProperties dbProperties = artifactoryHome.getDBProperties();
        DbLocksServiceImpl dbLocksService = new DbLocksServiceImpl(dbProperties);
        if (artifactoryHome.isHaConfigured()) {
            // Acquire lock for the node
            log.info("Attempting to acquire a high-availability init lock on key {}, with node-id {}",
                    ARTIFACTORY_ASYNC_CONVERTER_LOCK_KEY,
                    ownerId);
            try {
                dbLocksService
                        .acquireLock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(),
                                ARTIFACTORY_ASYNC_CONVERTER_LOCK_KEY,
                                ownerId, 1,
                                TimeUnit.HOURS);
                log.info("Acquired high-availability init lock on key {}", ARTIFACTORY_ASYNC_CONVERTER_LOCK_KEY);
            } catch (TimeoutException e) {
                String error = "Unable to acquire high-availability lock on category %s with key %s";
                log.error(error, e);
                return;
            }
        }

        // Run the converters
        DbType dbType = artifactoryHome.getDBProperties().getDbType();
        try {
            for (DBConverter converter : converters) {
                converter.convert(jdbcHelper, dbType);
            }
        } finally {
            // If HA remove the lock from the db
            if (artifactoryHome.isHaConfigured()) {
                dbLocksService.unlock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(),
                        ARTIFACTORY_ASYNC_CONVERTER_LOCK_KEY,
                        ownerId);
            }
        }
    }
}
