package org.artifactory.rest.common.validator;

import com.google.common.collect.Lists;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.repo.ReleaseBundlesRepositoryConfigurationImpl;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.util.UiRequestUtils;
import org.jdom2.Verifier;

import java.util.List;

import static org.apache.commons.lang.StringUtils.isBlank;
import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.artifactory.repo.system.SystemRepos.RELEASE_BUNDLE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.SUPPORT_BUNDLE_REPO_KEY;

public class RepositoryNameValidator {

    private static final int REPO_KEY_MAX_LENGTH = 64;
    private static final List<Character> forbiddenChars = Lists
            .newArrayList('/', '\\', ':', '|', '?', '*', '"', '<', '>');

    // TODO[ts]: Add here the blacklist check
    public static void validateRepoName(String repoKey, String repoType, CentralConfigService centralConfig)
            throws RepoConfigException {
        if (isBlank(repoKey)) {
            throw new RepoConfigException("Repository key cannot be empty", SC_BAD_REQUEST);
        }
        //Local repos limited at 64, remote repos append '-cache' to cache repos which is 6 chars.
        int maxLength = "remote".equals(repoType) ? (REPO_KEY_MAX_LENGTH - 6) : REPO_KEY_MAX_LENGTH;
        if (StringUtils.length(repoKey) > maxLength) {
            throw new RepoConfigException("Repository key exceeds maximum length (" + maxLength + ") chars.",
                    SC_BAD_REQUEST);
        }
        if (UiRequestUtils.isReservedName(repoKey)) {
            throw new RepoConfigException("Repository key '" + repoKey + "' is a reserved name", SC_BAD_REQUEST);
        }
        if (repoKey.equals(".") || repoKey.equals("..") || repoKey.equals("&")) {
            throw new RepoConfigException("Invalid Repository key", SC_BAD_REQUEST);
        }
        if (repoKey.chars().anyMatch(chr -> forbiddenChars.contains((char) chr))) {
            throw new RepoConfigException("Illegal Repository key : '/,\\,:,|,?,<,>,*,\"' is not allowed",
                    SC_BAD_REQUEST);
        }
        String error = Verifier.checkXMLName(repoKey);
        if (StringUtils.isNotBlank(error)) {
            throw new RepoConfigException("Repository key contains illegal character", SC_BAD_REQUEST);
        }

        if (!centralConfig.getMutableDescriptor().isKeyAvailable(repoKey)) {
            throw new RepoConfigException("Case insensitive repository key already exists", SC_BAD_REQUEST);
        }

        if (SUPPORT_BUNDLE_REPO_KEY.equals(repoKey)) {
            throw new RepoConfigException(
                    SUPPORT_BUNDLE_REPO_KEY + " is a reserved repository key",
                    SC_BAD_REQUEST);
        }

        if (RELEASE_BUNDLE_REPO_KEY.equals(repoKey) &&
                !ReleaseBundlesRepositoryConfigurationImpl.TYPE.equals(repoType)) {
            throw new RepoConfigException(
                    RELEASE_BUNDLE_REPO_KEY + " is a reserved repository key",
                    SC_BAD_REQUEST);
        }
    }
}
