package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.fetch;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.security.ArtifactoryPermission;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.TreeFilter;
import org.artifactory.util.CollectionUtils;

/**
 * This class is used to filter out unneeded descriptors in the various fetch strategies
 *
 * @author Omri Ziv
 */
public class DescriptorFilter {
    private final TreeFilter treeFilter;
    private final AuthorizationService authService;

    public DescriptorFilter(TreeFilter treeFilter) {
        this.treeFilter = treeFilter;
        this.authService = ContextHelper.get().getAuthorizationService();
    }

    public boolean accept(RepoDescriptor repoDescriptor) {
        return matchRepoFromList(repoDescriptor) &&
                matchRepoKey(repoDescriptor) &&
                matchPackageType(repoDescriptor) &&
                matchRepoType(repoDescriptor) &&
                (userCanReadAnywhereOnBuildRepo(authService, repoDescriptor)
                        || authService.userHasPermissionsOnRepositoryRoot(repoDescriptor.getKey()));
    }

    private boolean matchRepoType(RepoDescriptor repoDescriptor) {
        return CollectionUtils.isNullOrEmpty(treeFilter.getRepositoryTypes()) ||
                treeFilter.getRepositoryTypes().contains(RepositoryType.byDescriptor(repoDescriptor));
    }

    private boolean matchPackageType(RepoDescriptor repoDescriptor) {
        return CollectionUtils.isNullOrEmpty(treeFilter.getPackageTypes()) ||
                treeFilter.getPackageTypes().contains(repoDescriptor.getType());
    }

    private boolean matchRepoKey(RepoDescriptor repoDescriptor) {
        return StringUtils.isBlank(treeFilter.getByRepoKey()) ||
                StringUtils.contains(repoDescriptor.getKey(), treeFilter.getByRepoKey());
    }

    private boolean matchRepoFromList(RepoDescriptor repoDescriptor) {
        return CollectionUtils.isNullOrEmpty(treeFilter.getRepositoryKeys()) ||
                treeFilter.getRepositoryKeys().contains(repoDescriptor.getKey());
    }

    /**
     * Special case for the build info repo, since its improbable for admins to give users permissions on its root.
     * We allow any user that has read anywhere on the repo to see it, given that individual folders and files are
     * still filtered by permissions.
     */
    private boolean userCanReadAnywhereOnBuildRepo(AuthorizationService authService, RepoDescriptor repository) {
        return RepoType.BuildInfo.equals(repository.getType()) &&
                authService.hasBuildPermission(repository.getKey(), ArtifactoryPermission.READ);
    }
}
