package org.artifactory.ui.rest.service.admin.configuration.repositories.util.validator;

import com.google.common.collect.Maps;
import org.artifactory.addon.AddonsManager;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoLayoutBuilder;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeStub;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteCacheRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.BowerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.DockerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GenericTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GoTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.TypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualSelectedRepository;
import org.junit.Before;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.validation.constraints.NotNull;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_GO_REMOTE_METADATA_MISSED_RETRIEVAL_CACHE_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_MISSED_RETRIEVAL_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_RETRIEVAL_CACHE_PERIOD;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author AndreiK.
 */
@Test
public class RepoConfigValidatorTest {

    private RepoConfigValidator validator;

    @Before
    public void setup() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryHome.bind(new ArtifactoryHomeStub());
        InternalArtifactoryContext context = mock(InternalArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(false);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        validator = new RepoConfigValidator();
    }

    @Test(dataProvider = "provideDockerCachePeriodAndRepoType", description =
            "Actually sets the value when changed in the" +
                    " UI note the cast to DockerTypeSpecificConfigModel - this replicates the UI originated model")
    public void testRemoteRepoValidationOfRetrievalCachePeriodSecsOfCacheModel(Long cachePeriod, boolean isDockerModel,
            Long expected) {
        TypeSpecificConfigModel typeSpecificConfigModel;
        if (isDockerModel) {
            typeSpecificConfigModel = new DockerTypeSpecificConfigModel();
            ((DockerTypeSpecificConfigModel) typeSpecificConfigModel).setRetrievalCachePeriodSecs(cachePeriod);
        } else {
            typeSpecificConfigModel = new GenericTypeSpecificConfigModel();
        }
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                getAdvancedRepoConfigModelAndSetRetrievalCachePeriod(cachePeriod);
        RemoteRepositoryConfigModel model = getRemoteRepositoryConfigModel(remoteAdvancedRepositoryConfigModel,
                typeSpecificConfigModel);
        RemoteCacheRepositoryConfigModel cache = model.getAdvanced().getCache();

        validator.setRetrievalCachePeriodSecsOfCacheModel(model, cache);

        assertEquals((long) cache.getRetrievalCachePeriodSecs(), (long) expected);
    }

    @Test(dataProvider = "provideGoCachePeriodAndRepoType")
    public void testRemoteRepoValidationOfGoRetrievalCachePeriodSecsOfCacheModel(Long cachePeriod, boolean isGoModel,
            Long expected) {
        TypeSpecificConfigModel typeSpecificConfigModel;
        if (isGoModel) {
            typeSpecificConfigModel = new GoTypeSpecificConfigModel();
            ((GoTypeSpecificConfigModel) typeSpecificConfigModel)
                    .setMissedRetrievalCachePeriodSecs(cachePeriod);
        } else {
            typeSpecificConfigModel = new GenericTypeSpecificConfigModel();
        }
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                getAdvancedRepoConfigModelAndSetMissedRetrievalCachePeriod(cachePeriod);
        RemoteRepositoryConfigModel model = getRemoteRepositoryConfigModel(remoteAdvancedRepositoryConfigModel,
                typeSpecificConfigModel);
        RemoteCacheRepositoryConfigModel cache = model.getAdvanced().getCache();

        validator.setRetrievalCachePeriodSecsOfCacheModel(model, cache);

        assertEquals((long) cache.getMissedRetrievalCachePeriodSecs(), (long) expected);
    }

    @Test
    public void testMapRepoKeyToDescriptor() {
        String repo = "repo";
        VirtualSelectedRepository repository = new VirtualSelectedRepository();
        repository.setRepoName(repo);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RepoDescriptor repoDescriptor =
                configValidator.mapRepoKeyToDescriptor(repository, createMutableDescriptorMock(repo));
        assertEquals(repoDescriptor.getKey(), repo);
    }

    @Test
    public void testValidateSelectedReposInVirtualExist() throws RepoConfigException {
        String repo = "repo";
        List<VirtualSelectedRepository> repoKeys = new ArrayList<>();
        VirtualSelectedRepository repository = new VirtualSelectedRepository();
        repository.setRepoName(repo);
        repoKeys.add(repository);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        MutableCentralConfigDescriptor mutableDescriptorMock = createMutableDescriptorMock(repo);
        configValidator.validateSelectedReposInVirtualExist(repoKeys, mutableDescriptorMock);
        verify(mutableDescriptorMock, times(1)).getFederatedRepositoriesMap();
    }

    @Test
    public void testValidateMetadataTimeIsValidDefaultValues() throws RepoConfigException {
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        configValidator.validateMetadataTimeIsValid(cache);
    }

    @Test
    public void testValidateMetadataTimeIsValidWithZeroCachePeriod() throws RepoConfigException {
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        cache.setRetrievalCachePeriodSecs(0L);
        configValidator.validateMetadataTimeIsValid(cache);
    }

    @Test(expectedExceptions = RepoConfigException.class, expectedExceptionsMessageRegExp = ".*The Metadata retrieval" +
            " timeout not allowed to be bigger then Metadata Retrieval.*")
    public void testValidateMetadataTimeIsValidTimeoutBiggerThenCache() throws RepoConfigException {
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        cache.setRetrievalCachePeriodSecs(1L);
        configValidator.validateMetadataTimeIsValid(cache);
    }

    @Test(expectedExceptions = RepoConfigException.class, expectedExceptionsMessageRegExp = ".*The Metadata retrieval" +
            " timeout can't be zero or negative.*")
    public void testValidateMetadataTimeIsValidWhenTimeoutIllegal() throws RepoConfigException {
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        cache.setRetrievalCachePeriodSecs(1L);
        cache.setMetadataRetrievalTimeoutSecs(0L);
        configValidator.validateMetadataTimeIsValid(cache);
    }

    @Test
    public void testValidateRemoteValidFields() throws RepoConfigException {
        RemoteRepositoryConfigModel model = new RemoteRepositoryConfigModel();
        setGeneralRemoteConfig(model);
        RemoteAdvancedRepositoryConfigModel advanced = new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        cache.setMetadataRetrievalTimeoutSecs(10L);
        advanced.setCache(cache);
        model.setAdvanced(advanced);
        setBasicRemoteConfig(model);
        BowerTypeSpecificConfigModel mock = mock(BowerTypeSpecificConfigModel.class);
        when(mock.getRepoType()).thenReturn(RepoType.Bower);
        model.setTypeSpecific(mock);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        prepareValidation(configValidator);
        configValidator.validateRemote(model);
    }

    @Test(expectedExceptions = RepoConfigException.class, expectedExceptionsMessageRegExp = ".*The Metadata retrieval" +
            " timeout can't be zero or negative.*")
    public void testValidateRemoteNotValidMetadataTimeout() throws RepoConfigException {
        RemoteRepositoryConfigModel model = new RemoteRepositoryConfigModel();
        setGeneralRemoteConfig(model);
        RemoteAdvancedRepositoryConfigModel advanced = new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        cache.setMetadataRetrievalTimeoutSecs(0L);
        advanced.setCache(cache);
        model.setAdvanced(advanced);
        setBasicRemoteConfig(model);
        BowerTypeSpecificConfigModel mock = mock(BowerTypeSpecificConfigModel.class);
        when(mock.getRepoType()).thenReturn(RepoType.Bower);
        model.setTypeSpecific(mock);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        prepareValidation(configValidator);
        configValidator.validateRemote(model);
    }

    private void prepareValidation(RepoConfigValidator configValidator) {
        CentralConfigService configService = mock(CentralConfigService.class);
        configValidator.setCentralConfig(configService);
        CentralConfigDescriptorImpl centralConfigDescriptor = new CentralConfigDescriptorImpl();
        RepoLayout simpleRepoLayout = new RepoLayoutBuilder().name("simple-default")
                .artifactPathPattern("mm").distinctiveDescriptorPathPattern(false).build();
        centralConfigDescriptor.setRepoLayouts(List.of(simpleRepoLayout));
        when(configService.getDescriptor()).thenReturn(centralConfigDescriptor);
        AddonsManager addonsManager = mock(AddonsManager.class);
        configValidator.setAddonsManager(addonsManager);
        when(addonsManager.isEnterprisePlusInstalled()).thenReturn(true);
    }

    private void setGeneralRemoteConfig(RemoteRepositoryConfigModel model) {
        GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        general.setRepoKey("repo");
        model.setGeneral(general);
    }

    private void setBasicRemoteConfig(RemoteRepositoryConfigModel model) {
        RemoteBasicRepositoryConfigModel basic = mock(RemoteBasicRepositoryConfigModel.class);
        when(basic.getLayout()).thenReturn("simple-default");
        when(basic.getUrl()).thenReturn("http://localhost");
        model.setBasic(basic);
    }

    private MutableCentralConfigDescriptor createMutableDescriptorMock(String repoName) {
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        Map<String, FederatedRepoDescriptor> reposMap = Maps.newLinkedHashMap();
        FederatedRepoDescriptor federatedRepo = new FederatedRepoDescriptor();
        federatedRepo.setKey(repoName);
        reposMap.put(repoName, federatedRepo);
        when(descriptor.getFederatedRepositoriesMap()).thenReturn(reposMap);
        return descriptor;
    }

    @DataProvider
    private Object[][] provideDockerCachePeriodAndRepoType() {
        return new Object[][] {
                {null, true, DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD},
                {null, false, DEFAULT_RETRIEVAL_CACHE_PERIOD},
                {36000L, true, 36000L},
                {36000L, false, 36000L},
        };
    }

    @DataProvider
    private Object[][] provideGoCachePeriodAndRepoType() {
        return new Object[][] {
                {null, true, DEFAULT_GO_REMOTE_METADATA_MISSED_RETRIEVAL_CACHE_PERIOD},
                {null, false, DEFAULT_MISSED_RETRIEVAL_PERIOD},
                {36000L, true, 36000L},
                {36000L, false, 36000L},
        };
    }

    @NotNull
    private RemoteAdvancedRepositoryConfigModel getAdvancedRepoConfigModelAndSetRetrievalCachePeriod(Long cachePeriod) {
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel remoteCacheRepositoryConfigModel = new RemoteCacheRepositoryConfigModel();
        remoteAdvancedRepositoryConfigModel.setCache(remoteCacheRepositoryConfigModel);
        remoteCacheRepositoryConfigModel.setRetrievalCachePeriodSecs(cachePeriod);
        return remoteAdvancedRepositoryConfigModel;
    }

    @NotNull
    private RemoteAdvancedRepositoryConfigModel getAdvancedRepoConfigModelAndSetMissedRetrievalCachePeriod(
            Long cachePeriod) {
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel remoteCacheRepositoryConfigModel = new RemoteCacheRepositoryConfigModel();
        remoteAdvancedRepositoryConfigModel.setCache(remoteCacheRepositoryConfigModel);
        remoteCacheRepositoryConfigModel.setMissedRetrievalCachePeriodSecs(cachePeriod);
        return remoteAdvancedRepositoryConfigModel;
    }

    @NotNull
    private RemoteRepositoryConfigModel getRemoteRepositoryConfigModel(
            RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel,
            TypeSpecificConfigModel typeSpecificConfigModel) {
        RemoteRepositoryConfigModel remoteRepositoryConfigModel = new RemoteRepositoryConfigModel();
        remoteRepositoryConfigModel.setBasic(new RemoteBasicRepositoryConfigModel());
        remoteRepositoryConfigModel.setGeneral(new GeneralRepositoryConfigModel());
        remoteRepositoryConfigModel.setAdvanced(remoteAdvancedRepositoryConfigModel);
        remoteRepositoryConfigModel.setTypeSpecific(typeSpecificConfigModel);
        return remoteRepositoryConfigModel;
    }
}